{*******************************************************}
{          Copyright (c) 2000 Vlad Korshunov            }
{                   IPH tools project                   }
{                      File parser                      }
{*******************************************************}
unit Parser;
interface
uses windows, sysutils, classes;

type
  TParserState = (stNone, stTreeItem, stParamSource, stViewGroup, stGroupParam,
    stComment, stResult, stResultParams, stParamValue);
  TParserStates = set of TParserState;

const
  StartTokens: array[TParserState] of String =
    ('', '[BeginTree]', '[BeginParams]', '[BeginSessionView]', '[BeginViewParams]',
     '[BeginComment]', '[BeginResults]', '[BeginResultParams]', '[BeginValue]');
  StopTokens:  array[TParserState] of String =
    ('', '[EndTree]', '[EndParams]', '[EndSessionView]', '[EndViewParams]',
     '[EndComment]', '[EndResults]', '[EndResultParams]', '[EndValue]');

type
  EParser = class(Exception);

  TParser = class
  private
    FState: TParserState;
    FStateStack: TList;
    function GetCurToken: String;
    function CheckToken: Boolean;
    procedure ChangeState(NewState: TParserState);
    procedure SkipBlanks;
    procedure PushState(State: TParserState);
    function PopState: TParserState;
  protected
    procedure Next; virtual; abstract;
    function EOF: Boolean; virtual; abstract;
    function ExpectedStates: TParserStates; virtual; abstract;
    procedure DoState; virtual; abstract;
    procedure DoStateChanged(Activated: Boolean); virtual;
    function GetCurLine: String; virtual; abstract;
  public
    constructor Create;
    destructor Destroy; override;
    procedure Parse;
    property CurLine: String read GetCurLine;
    property CurToken: String read GetCurToken;
    property State: TParserState read FState;
  end;

  TTreeParser = class(TParser)
  protected
    function GetLevel: Integer;
  end;

  TStreamParser = class(TTreeParser)
  private
    FStrings: TStrings;
    FIndex: Integer;
  protected
    procedure Next; override;
    function EOF: Boolean; override;
    function GetCurLine: String; override;
  public
    constructor Create(Stream: TStream);
    destructor Destroy; override;
  end;

procedure WriteStateBegin(Stream: TStream; State: TParserState);
procedure WriteStateEnd(Stream: TStream; State: TParserState);

implementation
uses iphconsts, utils;

{ TParser }

procedure TParser.ChangeState(NewState: TParserState);

  function GetStateNames(States: TParserStates): string;
  var
    i: TParserState;
  begin
    result:='';
    for i:=Low(TParserState) to High(TParserState) do
    if i in States then
    begin
      if result<>'' then result:=result+', ';
      result:=result+StartTokens[i];
    end;
  end;
  
begin
  if not (NewState in ExpectedStates) then
    Raise EParser.CreateFmt(sNotExpectToken, [StartTokens[NewState],
      GetStateNames(ExpectedStates)]);
  FState:=NewState;
end;

function TParser.CheckToken: Boolean;
var
  i : TParserState;
begin
  Result:=false;
  for i:=succ(Low(TParserState)) to High(TParserState) do
    if AnsiCompareText(CurToken, StartTokens[i])=0 then
    begin
      Result:=true;
      PushState(State);
      ChangeState(i);
      Next;
      DoStateChanged(True);
      break;
    end else
    if AnsiCompareText(CurToken, StopTokens[i])=0 then
    begin
      result:=True;
      DoStateChanged(False);
      Next;
      FState:=PopState;
      break;
    end;
end;

constructor TParser.Create;
begin
  inherited Create;
  FState:=stNone;
  FStateStack:=TList.Create;
end;

destructor TParser.Destroy;
begin
  FStateStack.Free;
  inherited;
end;

procedure TParser.DoStateChanged(Activated: Boolean);
begin
 // do nothing
end;

function TParser.GetCurToken: String;
begin
  result:=AnsiUpperCase(Trim(CurLine));
end;

procedure TParser.Parse;
begin
  SkipBlanks;
  while not EOF do
  begin
    if not CheckToken then
    begin
      DoState;
      Next;
    end;
    SkipBlanks;
  end;
end;

function TParser.PopState: TParserState;
begin
  if FStateStack.Count>0 then
  begin
    result:=TParserState(FStateStack[FStateStack.Count-1]);
    FStateStack.Delete(FStateStack.Count-1);
  end else
    Raise EParser.Create(sParserStackEmpty);
end;

procedure TParser.PushState(State: TParserState);
begin
  FStateStack.Add(Pointer(State));
end;

procedure TParser.SkipBlanks;
var
  s: String;
begin
  while not EOF do
  begin
    s:=Trim(CurLine);
    if (s <> '') and (Pos(tkComment, s)<>1) then break;
    Next;
  end;
end;

{ TStreamParser }

constructor TStreamParser.Create(Stream: TStream);
var
  Size: Integer;
  S: string;
 i: integer;
begin
  inherited Create;
  FStrings:=TStringList.Create;
  With FStrings do
  begin
    BeginUpdate;
    try
      Size := Stream.Size - Stream.Position;
      SetString(S, nil, Size);
      Stream.Read(Pointer(S)^, Size);
      for i:=1 to Size do
        if s[i] = #0 then s[i] := #13;
      Text:=S;
    finally
      EndUpdate;
    end;
  end;
  FIndex:=0;
end;

destructor TStreamParser.Destroy;
begin
  FStrings.Free;
  inherited;
end;

function TStreamParser.EOF: Boolean;
begin
  result:=FIndex>=FStrings.Count-1;
end;

function TStreamParser.GetCurLine: String;
begin
  result:=FStrings[FIndex];
end;

procedure TStreamParser.Next;
begin
  inc(FIndex);
end;

{ TTreeParser }

function TTreeParser.GetLevel: Integer;
var
  i: integer;
  FCurLine: String;
begin
  FCurLine:=CurLine;
  result:=0;
  for i:=1 to Length(FCurLine) do
    if not (FCurLine[i] in [TabChar]) then
    begin
      result:=i-1;
      break;
    end;
end;


procedure WriteStateBegin(Stream: TStream; State: TParserState);
begin
  StreamWriteStr(Stream, StartTokens[State]+EndOfLine);
end;

procedure WriteStateEnd(Stream: TStream; State: TParserState);
begin
  StreamWriteStr(Stream, StopTokens[State]+EndOfLine+EndOfLine);
end;

end.
