{*******************************************************}
{          Copyright (c) 2000 Vlad Korshunov            }
{                   IPH tools project                   }
{                  Base probe classes                   }
{*******************************************************}

unit probes;
interface
uses windows, messages, sysutils, classes,
  ProbeParams, PluginAPI {$IFDEF LOG}, LogFile {$ENDIF};

{
Common rules
  1. Do not any action in Create, because probe may be created by template
  editor to retrieve info about params
  2. StartProbe must return immediately
  3. StopProbe must not return until probe is really finished, after this probe
    will be immediately destroyed
}
type
  TCustomProbe = class;

  TProbeClass = class of TCustomProbe;

  TProbeNotify = procedure(Probe: TCustomProbe) of object;

  TCustomProbe = class
  private
    FHWND: THandle;
    FParams: TParams;
    FTerminating: Boolean;
    FActive: Boolean;
    FNotifyByMessage: Boolean;
    FNotify: TProbeNotify;
    FOnIdle: TProbeNotify;
    procedure WndProc(var Msg : TMessage);
  protected
    procedure CheckRequiredParams; virtual;
    procedure NotifyThread(Success: Boolean);
    procedure DoOnProbe(Finished: Boolean);
    procedure DoOnError(ErrorCode: Integer; const ErrorMsg: string;
      Finished: Boolean = True);
    procedure AddSupportedParams(var Params: TParams); virtual;
    procedure InternalStartProbe; virtual; abstract;
    procedure InternalStopProbe; virtual;
    procedure InternalWndProc(var Msg: TMessage); virtual;
    function GetProbeName: string; virtual;               // for custom probes
    function GetProbeDescription: String; virtual;
    procedure SafePostMessage(HWND: HWND; Msg, WParam, LParam: Integer);
    procedure Idle;
  public
    class function GetName: string; virtual; abstract;    // for predefined probes
    class function GetDescription: String; virtual; abstract;
    constructor Create; virtual;
    destructor destroy; override;
    procedure StartProbe;
    procedure StopProbe;
    property NotifyByMessage: Boolean read FNotifyByMessage write FNotifyByMessage;
    property Name: string read GetProbeName;
    property Params: TParams read FParams write FParams;
    property Terminating: Boolean read FTerminating;
    property Active: Boolean read FActive;
    property HWND: THandle read FHWND;
    property Description: String read GetProbeDescription;
    property Notify: TProbeNotify read FNotify write FNotify;
    property OnIdle: TProbeNotify read FOnIdle write FOnIdle;
  end;

  TPlugin = class
  private
    FPluginPath: String;
    FDLLHandle: Integer;
    FSupportedProbes: TStrings;
    // API
    FGetProbeInfo: TGetProbeInfo;
    FEnumProbeParams: TEnumProbeParams;
    FStartProbe: TStartProbe;
    FStopProbe: TStopProbe;
    FGetCurrentValues: TGetCurrentValues;
    FEnumProbeNames: TEnumProbeNames;
    //
    procedure Load;
    procedure Unload;
    function GetSupportedProbes: TStrings;
  public
    constructor Create(const PluginPath: String);
    destructor Destroy; override;
    function GetProbeInfo(var ProbeName: String): String;
    procedure GetProbeParams(const ProbeName: String; Params: TParams);
    function StartProbe(const ProbeName: String; Params: TParams;
      Msg: Integer; HWND: HWND): Integer;
    procedure GetCurrentValues(ProbeHandle: Integer; Params: TParams);
    procedure StopProbe(ProbeHandle: Integer);
    property SupportedProbes: TStrings read GetSupportedProbes;
  end;

  TPluginProbe = class(TCustomProbe)
  private
    FPlugin: TPlugin;
    FProbeHandle: Integer;
    FProbeName: String;
    procedure ProbeNotify;
  protected
    procedure AddSupportedParams(var Params: TParams); override;
    procedure InternalStartProbe; override;
    procedure InternalStopProbe; override;
    procedure InternalWndProc(var Msg: TMessage); override;
    function GetProbeName: string; override;
    function GetProbeDescription: String; override;
  public
    class function GetName: string; override;
    class function GetDescription: String; override;
    constructor CreateFromPlugin(Plugin: TPlugin; const ProbeName: String);
    destructor destroy; override;
  end;

  TTimeredProbe = class(TCustomProbe)
  private
    FTimer: UINT;
  protected
    procedure InternalStartProbe; override;
    procedure InternalStopProbe; override;
    procedure AddSupportedParams(var Params: TParams); override;
    procedure InternalWndProc(var Msg: TMessage); override;
    procedure StopTimer;
  public
  end;

  TActionProbe = class(TCustomProbe)
  protected
    procedure InternalStartProbe; override;
    procedure InternalWndProc(var Msg: TMessage); override;
    function InternalAction: Boolean; virtual; abstract;
    function NextAction: Boolean; virtual; abstract;
  end;

  TIteratorProbe = class(TActionProbe)
  protected
  end;

  TRangeIterator = class(TIteratorProbe)
  private
    FLoValueName, FHiValueName, FCurValueName: string;
    FLoValue: Integer;
    FHiValue: Integer;
    FCurValue: Integer;
  protected
    procedure CheckRequiredParams; override;
    procedure AddSupportedParams(var Params: TParams); override;
    function InternalAction: Boolean; override;
    function NextAction: Boolean; override;
    procedure GetParamNames(var LoValueName, HiValueName, CurValueName: String;
      var ValueType: TFieldType); virtual; abstract;
    function OrdinalForm(ComplexValue: Integer): Integer; virtual;
    function ComplexForm(OrdinalValue: Integer): Integer; virtual;
  end;

function GetProbeByName(const ProbeName: String): TCustomProbe;
procedure GetAvailableProbes(Probes: TStrings);
procedure RegisterProbes(ProbeClasses: array of TProbeClass);
procedure CreatePseudoParams(Params: TParams);

const
  PluginsDirectory = 'Plugins';
  PluginsMask = '*.dll';

  WM_PROBE_BASE         = wm_user+$400;
  WM_PROBE_EVENT        = WM_PROBE_BASE;
  WM_PLUGIN_PROBE       = WM_PROBE_BASE+1;
  WM_NEXT_ACTION        = WM_PROBE_BASE+2;
  WM_ACTION             = WM_PROBE_BASE+3;

  MaxMessages           = 30;

  pnProbeFinished = 'ProbeFinished';
  pnProbeSuccess  = 'ProbeSuccess';

implementation
uses iphconsts, utils;

procedure CreatePseudoParams(Params: TParams);
var
  Param: TParam;
begin
  Param:=Params.FindParam(pnProbeFinished);
  if Param = nil then
    Params.CreateParam(ftBoolean, pnProbeFinished, ptOutput, []);
  Param:=Params.FindParam(pnProbeSuccess);
  if Param = nil then
    Params.CreateParam(ftBoolean, pnProbeSuccess, ptOutput, []);
end;

function ProbeWndProc(HWND: HWND; Msg: Integer; wParam : WPARAM;
  lParam : LPARAM): Integer; stdcall;
var
  Probe: TCustomProbe;
  Message : TMessage;
begin
  Probe := TCustomProbe(GetWindowLong(HWND, 0));
  if Probe = nil then
    Result := DefWindowProc(HWND, Msg, wParam, lParam) else
  begin
    Message.Msg    := Msg;
    Message.wParam := wParam;
    Message.lParam := lParam;
    Probe.WndProc(Message);
    Result := Message.Result;
  end;
end;

var
  ProbeWindowClass: TWndClass = (
    style         : 0;
    lpfnWndProc   : @ProbeWndProc;
    cbClsExtra    : 0;
    cbWndExtra    : SizeOf(Pointer);
    hInstance     : 0;
    hIcon         : 0;
    hCursor       : 0;
    hbrBackground : 0;
    lpszMenuName  : nil;
    lpszClassName : 'IPHTools_Probe_Class');

function AllocateHWnd(Probe : TCustomProbe): HWND;
var
    TempClass       : TWndClass;
    ClassRegistered : Boolean;
begin
  ProbeWindowClass.hInstance := HInstance;
  ClassRegistered := GetClassInfo(HInstance, ProbeWindowClass.lpszClassName,
    TempClass);
  if not ClassRegistered then
  begin
    Result := Windows.RegisterClass(ProbeWindowClass);
    if Result = 0 then Exit;
  end;
  Result := CreateWindowEx(WS_EX_TOOLWINDOW, ProbeWindowClass.lpszClassName,
    '', WS_POPUP, 0, 0, 0, 0, 0, 0, HInstance, nil);
  if (Result <> 0) and Assigned(Probe) then
    SetWindowLong(Result, 0, Integer(Probe));
end;

procedure DeallocateHWnd(Wnd: HWND);
begin
  DestroyWindow(Wnd);
end;

const
  FRegisteredProbes: TStringList = nil;
  FPlugins: TStringList = nil;

procedure InitPlugins;
var
  Found: Integer;
  SearchRec: TSearchRec;
  Plugin: TPlugin;
  Dir: String;
begin
  if FPlugins=nil then
  begin
    Dir:=IncludeTrailingBackSlash(IncludeTrailingBackSlash(ExtractFilePath(
      GetModuleName(HInstance)))+PluginsDirectory);
    FPlugins:=TStringList.Create;
    Found := FindFirst(Dir+PluginsMask, faAnyFile, SearchRec);
    try
      while Found = 0 do
      begin
        try
          Plugin:=TPlugin.Create(Dir+SearchRec.Name);
          FPlugins.AddObject(Dir+SearchRec.Name, Plugin);
        except
          On E: Exception do MessageBox(0, PChar(E.Message), '', 0);
        end;
        Found := FindNext(SearchRec);
      end;
    finally
      FindClose(SearchRec);
    end;
  end;
end;

procedure RegisterProbes(ProbeClasses: array of TProbeClass);
var
  i: integer;
begin
  if FRegisteredProbes = nil then FRegisteredProbes:=TStringList.Create;
  for i:=Low(ProbeClasses) to High(ProbeClasses) do
    FRegisteredProbes.AddObject(ProbeClasses[i].GetName, Pointer(ProbeClasses[i]));
end;

function GetProbeByName(const ProbeName: String): TCustomProbe;
var
  i: integer;
begin
  result:=nil;
  if FRegisteredProbes<>nil then
    for i:=0 to FRegisteredProbes.Count-1 do
      if AnsiUpperCase(TProbeClass(FRegisteredProbes.Objects[i]).GetName) =
        AnsiUpperCase(ProbeName) then
      begin
        result:=TProbeClass(FRegisteredProbes.Objects[i]).Create;
        break;
      end;
  if result = nil then
  begin
    InitPlugins;
    for i:=0 to FPlugins.Count-1 do
      if TPlugin(FPlugins.Objects[i]).SupportedProbes.IndexOf(ProbeName)<>-1 then
      begin
        result:=TPluginProbe.CreateFromPlugin(TPlugin(FPlugins.Objects[i]), ProbeName);
        Break;
      end;
  end;
end;

procedure GetAvailableProbes(Probes: TStrings);
var
  i, j: Integer;
begin
  if FRegisteredProbes<>nil then
    Probes.Assign(FRegisteredProbes);
  InitPlugins;
  for i:=0 to FPlugins.Count-1 do
    for j:=0 to TPlugin(FPlugins.Objects[i]).SupportedProbes.Count-1 do
      Probes.Add(TPlugin(FPlugins.Objects[i]).SupportedProbes[j]);
end;

{ TCustomProbe }
procedure TCustomProbe.AddSupportedParams(var Params: TParams);
begin
  // create default error params
  Params.CreateParam(ftInteger, 'ErrorCode', ptOutput, []);
  Params.CreateParam(ftString, 'ErrorMsg', ptOutput, []);
end;

procedure TCustomProbe.CheckRequiredParams;
var
  i: Integer;
begin
  for i:=0 to FParams.Count-1 do
  begin
    if (FParams[i].ParamType=ptInput) and (poRequired in FParams[i].ParamOptions)
      and FParams[i].IsNull then
      raise exception.CreateFmt(sProbeParamEmpty, [FParams[i].Name]);
  end;
end;

constructor TCustomProbe.Create;
begin
  inherited Create;
  FParams:=TParams.Create;
  AddSupportedParams(FParams);
  FNotifyByMessage:=True;
end;

destructor TCustomProbe.destroy;
begin
  if HWND<>0 then DeallocateHwnd(HWND);
  FParams.Free;
  inherited;
end;

procedure TCustomProbe.DoOnError(ErrorCode: Integer; const ErrorMsg: string;
  Finished: Boolean = True);
begin
  if ErrorCode<>0 then
    Params.ParamByName('ErrorCode').AsInteger:=ErrorCode else
    Params.ParamValues['ErrorCode']:=NULL;
  Params.ParamByName('ErrorMsg').AsString:=ErrorMsg;
  FActive:=not Finished;
  NotifyThread(False);
end;

procedure TCustomProbe.DoOnProbe(Finished: Boolean);
begin
  if Finished then FActive:=False;
  NotifyThread(True);
end;

function TCustomProbe.GetProbeDescription: String;
begin
  result:=GetDescription;
end;

function TCustomProbe.GetProbeName: string;
begin
  result:=GetName;
end;

procedure TCustomProbe.Idle;
begin
  if Assigned(FOnIdle) then FOnIdle(Self);
end;

procedure TCustomProbe.InternalStopProbe;
begin
  while FActive do Idle;
end;

procedure TCustomProbe.InternalWndProc(var Msg: TMessage);
begin
end;

procedure TCustomProbe.NotifyThread(Success: Boolean);
var
  ChildParams: TParams;
begin
  if FNotifyByMessage then
  begin
    ChildParams:=TParams.Create;      
    ChildParams.Assign(Params);
  end else
    ChildParams:=Params;
  CreatePseudoParams(ChildParams);
  ChildParams.ParamByName(pnProbeFinished).AsBoolean:=Not FActive;
  ChildParams.ParamByName(pnProbeSuccess).AsBoolean:=Success;
  {$IFDEF LOG} LogError(Name+' Success:'+ChildParams.ParamByName(pnProbeSuccess).AsString+
    '. Finished: '+ChildParams.ParamByName(pnProbeFinished).AsString); {$ENDIF}

  if FNotifyByMessage then
    SafePostMessage(0, WM_PROBE_EVENT, Integer(Pointer(Self)), Integer(Pointer(ChildParams)))
  else
    if Assigned(FNotify) then FNotify(Self);
end;

procedure TCustomProbe.SafePostMessage(HWND: HWND; Msg, WParam,
  LParam: Integer);
var
  Message: TMSG;
begin
  while PeekMessage(Message, HWND, Msg, Msg, PM_NOREMOVE) do Idle;
  if not PostMessage(Hwnd, Msg, WParam, LParam) then
    RaiseLastWin32Error('PostMessage');
end;

procedure TCustomProbe.StartProbe;
begin
  try
    CheckRequiredParams;
    FHWND:=AllocateHwnd(Self);
    FActive:=True;
    InternalStartProbe;
    FTerminating:=False;
  except
    On E: Exception do
    begin
      FActive:=False;
      Raise;
    end;
  end;
end;

procedure TCustomProbe.StopProbe;
begin
  if FActive then
  begin
    FTerminating:=True;
    InternalStopProbe;
  end;
end;

procedure TCustomProbe.WndProc(var Msg: TMessage);
begin
  if not FActive then Exit;
  if Terminating then FActive:=False;
  try
    InternalWndProc(Msg);
  except
    On E: Exception do DoOnError(0, E.Message);
  end;
end;

{ TActionProbe }

procedure TActionProbe.InternalStartProbe;
begin
  SafePostMessage(HWND, WM_ACTION, 0, 0);
end;

procedure TActionProbe.InternalWndProc(var Msg: TMessage);
begin
  Case Msg.Msg of
    WM_ACTION:
      begin
        {$IFDEF LOG} LogError(Name+' WM_ACTION'); {$ENDIF}
        if InternalAction then
          SafePostMessage(HWND, WM_NEXT_ACTION, 0, 0);
      end;
    WM_NEXT_ACTION:
      begin
        {$IFDEF LOG} LogError(Name+' NEXT_ACTION'); {$ENDIF}
        if NextAction then
          SafePostMessage(HWND, WM_NEXT_ACTION, 0, 0);
      end;
    else inherited
  end;
end;

{ TRangeIterator }

procedure TRangeIterator.AddSupportedParams(var Params: TParams);
var
  FValueType: TFieldType;
begin
  inherited;
  GetParamNames(FLoValueName, FHiValueName, FCurValueName,
    FValueType);
  Params.CreateParam(FValueType, FLoValueName, ptInput, [poRequired]);
  Params.CreateParam(FValueType, FHiValueName, ptInput, [poRequired]);
  Params.CreateParam(FValueType, FCurValueName, ptOutput, []);
end;

procedure TRangeIterator.CheckRequiredParams;
begin
  inherited;
  if OrdinalForm(Params.ParamValues[FLoValueName])>OrdinalForm(Params.ParamValues[FHiValueName]) then
    Raise Exception.CreateFmt(sParamMustBeNotLess, [FHiValueName, FLoValueName]);
end;

function TRangeIterator.ComplexForm(OrdinalValue: Integer): Integer;
begin
  result:=OrdinalValue;
end;

function TRangeIterator.InternalAction: Boolean;
begin
  FLoValue:=OrdinalForm(Params.ParamValues[FLoValueName]);
  FHiValue:=OrdinalForm(Params.ParamValues[FHiValueName]);
  FCurValue:=FLoValue;
  Result:=True;
end;

function TRangeIterator.NextAction: Boolean;
begin
  result:=False;
  if FCurValue<=FHiValue then
  begin
    Params.ParamValues[FCurValueName]:=ComplexForm(FCurValue);
    Result:=not (FCurValue = FHiValue);
    DoOnProbe(not result);
    inc(FCurValue);
  end;
end;

function TRangeIterator.OrdinalForm(ComplexValue: Integer): Integer;
begin
  result:=ComplexValue;
end;

{ TTimeredProbe }

procedure TTimeredProbe.InternalWndProc(var Msg: TMessage);
begin
  case msg.Msg of
    WM_TIMER:
      begin
        StopTimer;
        DoOnError(0, 'Timeout');
      end;
    else inherited;
  end;
end;

procedure TTimeredProbe.StopTimer;
begin
  if FTimer<>0 then KillTimer(HWND, 1);
  FTimer:=0;
end;

procedure TTimeredProbe.InternalStartProbe;
begin
  FTimer:=SetTimer(HWND, 1, Params.ParamByName('Timeout').AsInteger, nil);
end;

procedure TTimeredProbe.AddSupportedParams(var Params: TParams);
begin
  inherited;
  Params.CreateParam(ftInteger, 'Timeout', ptInput, []).Value:=5000;
end;

procedure TTimeredProbe.InternalStopProbe;
begin
  StopTimer;
  inherited;
end;

{ TPlugin }

constructor TPlugin.Create(const PluginPath: String);
begin
  inherited Create;
  FPluginPath:=PluginPath;
  Load;
end;

destructor TPlugin.Destroy;
begin
  Unload;
  inherited;
end;

function cbSetValue(ParamName: PChar; Value: PChar; Data: Pointer): Boolean; stdcall;
var
  Param: TParam;
begin
  result:=True;
  if (Data = nil) then
    Raise exception.Create(sInvalidData);
  Param:=TParams(Data).FindParam(ParamName);
  if Param<>nil then Param.AsString:=Value;
end;

procedure TPlugin.GetCurrentValues(ProbeHandle: Integer; Params: TParams);
begin
  FGetCurrentValues(ProbeHandle, @cbSetValue, Params);
end;

function TPlugin.GetProbeInfo(var ProbeName: String): String;
var
  ReqSize: Integer;
begin
  Result:='';
  reqSize:=0;
  if not FGetProbeInfo(PChar(ProbeName), @Result[1], ReqSize) then
  begin
    SetLength(Result, ReqSize);
    FGetProbeInfo(PChar(ProbeName), @Result[1], ReqSize);
  end;
end;

function ProbeParamCallback(ParamName: PChar; ParamType: Integer;
  Input: Boolean; Required: Boolean; DefaultValue: PChar; Data: Pointer): Boolean; stdcall;
var
  pType: TParamType;
  Options: TParamOptions;
  Value: String;
  Param: TParam;
begin
  result:=True;
  if (Data = nil) then
    Raise exception.Create(sInvalidData);
  with TParams(Data) do
  begin
    if Input then pType:=ptInput else pType:=ptOutput;
    Options:=[];
    if Required then Include(Options, poRequired);
    Value:=DefaultValue;
    Param:=FindParam(ParamName);
    if Param=nil then Param:=CreateParam(TFieldType(ParamType), ParamName, pType, Options);
    if Value<>'' then Param.AsString:=DefaultValue
  end;
end;

procedure TPlugin.GetProbeParams(const ProbeName: String; Params: TParams);
begin
  FEnumProbeParams(PChar(ProbeName), ProbeParamCallback, Params);
end;

function ProbeNameCallback(ProbeName: PChar; Data: pointer): Boolean; stdcall;
begin
  result:=True;
  if (Data = nil) then
    Raise exception.Create(sInvalidData);
  TStrings(Data).Add(ProbeName);
end;

function TPlugin.GetSupportedProbes: TStrings;
begin
  if FSupportedProbes = nil then
  begin
    FSupportedProbes:=TStringList.Create;
    FEnumProbeNames(ProbeNameCallback, Pointer(FSupportedProbes))
  end;
  result:=FSupportedProbes;
end;

procedure TPlugin.Load;
begin
  if FDLLHandle=0 then
  begin
    FDLLHandle:=SafeLoadLibrary(FPluginPath);
    if FDLLHandle = 0 then
      raise Exception.CreateFmt(sErrorLoadingPlugin,
        [FPluginPath, SysErrorMessage(GetLastError)]);
    BindProcs(FDLLHandle,
      [@@FGetProbeInfo, @@FEnumProbeParams, @@FStartProbe, @@FStopProbe, @@FGetCurrentValues],
      [efGetProbeInfo, efEnumProbeParams, efStartProbe, efStopProbe, efGetCurrentValues]);
    FEnumProbeNames:=GetProcAddress(FDLLHandle, efEnumProbeNames);
  end;
end;

function cbGetValue(ParamName: PChar; Data: Pointer; Value: PChar;
  var BufSize: Integer): Boolean; stdcall;
var
  Param: TParam;
begin
  if (Data = nil) then
    Raise exception.Create(sInvalidData);
  Param:=TParams(Data).FindParam(ParamName);
  if Param<>nil then result:=CopyStringToBuf(Param.AsString, Value, BufSize) else
    Raise Exception.CreateFmt(SParameterNotFound, [ParamName]);
end;

function TPlugin.StartProbe(const ProbeName: String; Params: TParams;
  Msg: Integer; HWND: HWND): Integer;
begin
  result:=FStartProbe(PChar(ProbeName), cbGetValue, Msg, HWND, Pointer(Params));
end;

procedure TPlugin.StopProbe(ProbeHandle: Integer);
begin
  FStopProbe(ProbeHandle);
end;

procedure TPlugin.Unload;
begin
  if FDLLHandle<>0 then
  begin
    FreeLibrary(FDLLHandle);
    FDLLHandle:=0;
  end;
end;

{ TPluginProbe }

procedure TPluginProbe.AddSupportedParams(var Params: TParams);
begin
  inherited;
  FPlugin.GetProbeParams(FProbeName, Params);
end;

constructor TPluginProbe.CreateFromPlugin(Plugin: TPlugin; const ProbeName: String);
begin
  FPlugin:=Plugin;
  FProbeName:=ProbeName;
  inherited Create;
end;

destructor TPluginProbe.destroy;
begin
  inherited;
end;

class function TPluginProbe.GetDescription: String;
begin
  result:='';
end;

class function TPluginProbe.GetName: string;
begin
  Raise exception.Create('Plugin probe does not have name');
end;

function TPluginProbe.GetProbeDescription: String;
begin
  result:=FPlugin.GetProbeInfo(FProbeName);
end;

function TPluginProbe.GetProbeName: string;
begin
  result:=FProbeName;
end;

procedure TPluginProbe.InternalStartProbe;
begin
  FProbeHandle:=FPlugin.StartProbe(FProbeName, Params, WM_PLUGIN_PROBE, HWND);
end;

procedure TPluginProbe.InternalStopProbe;
begin
  FPlugin.StopProbe(FProbeHandle);
  inherited;
end;

procedure TPluginProbe.InternalWndProc(var Msg: TMessage);
begin
  if (Msg.Msg = WM_PLUGIN_PROBE) and (Msg.WParam = FProbeHandle) then
  begin
    ProbeNotify;
  end else
    inherited;
end;

procedure Finit;
var
  i: integer;
begin
  if FRegisteredProbes<>nil then FRegisteredProbes.Free;
  if FPlugins<>nil then
  begin
    for i:=FPlugins.Count-1 downto 0 do
    begin
      FPlugins.Objects[i].Free;
      FPlugins.Delete(i);
    end;
    FPlugins.Free;
  end;
end;

procedure TPluginProbe.ProbeNotify;
var
  _Success, _Finished: Boolean;
begin
  CreatePseudoParams(Params);
  FPlugin.GetCurrentValues(FProbeHandle, Params);
  _Success:=Params.ParamByName(pnProbeSuccess).AsBoolean;
  _Finished:=Params.ParamByName(pnProbeFinished).AsBoolean;
  FActive:= _Success and not _Finished;
  NotifyThread(_Success);
end;

initialization
finalization
  Finit;
end.


